/*
Copyright  2005 The Regents of the University of California.
    All Rights Reserved.

    Permission to use, copy, modify, and distribute this software and its
    documentation for educational, research and non-profit purposes,
    without fee, and without a written agreement is hereby granted,
    provided that the above copyright notice, this paragraph and the
    following three paragraphs appear in all copies.

    Permission to incorporate this software into commercial products may
    be obtained by contacting the University of California. For
    information about obtaining such a license contact:

    Chrisanna Waldrop
    Copyright Officer
    805-893-7773
    waldrop@research.ucsb.edu

    This software program and documentation are copyrighted by The Regents
    of the University of California. The software program and
    documentation are supplied "as is", without any accompanying services
    from The Regents. The Regents does not warrant that the operation of
    the program will be uninterrupted or error-free. The end-user
    understands that the program was developed for research purposes and
    is advised not to rely exclusively on the program for any reason.

    IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
    FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
    INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
    ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
    CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
    BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
    MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

#pragma warning (disable: 4786)

#include <IGMP.h>

#include <ws2tcpip.h>

MD_Igmp::MD_Igmp( struct sockaddr_in& group )
 : m_socket( IPPROTO_IGMP ),
   m_group( group ),
   m_response( None )
{
	m_socket.Bind( IPPROTO_IGMP );

	m_socket.SetIoctlForIGMP();
}


void MD_Igmp::ReceiveQuery( int timeoutMs, bool waitForMultipleQuery )
{
    sockaddr_in source;

    char  buffer[4096];

    memset( &source, 0, sizeof( sockaddr_in ) );

    int waitInterval = timeoutMs;

    int timestamp = GetTickCount();

    while( waitInterval > 0 )
    {
        if( m_socket.IsDataAvailableForRead( waitInterval ) )
        {
            int bytes = m_socket.ReceiveFrom( buffer, 4096, &source );

            IPHeader* ipHeader = (IPHeader*)buffer;

            // Skip ahead to the ICMP header within the IP packet
            unsigned short len = ipHeader->ip_hl * 4;
            IGMPHeader* igmphdr = (IGMPHeader*)(buffer + len);

            //TBD : Do size and checksum verification
            if (igmphdr->igmp_type == IGMP_MEMBERSHIP_QUERY ) 
            {
                if( igmphdr->igmp_group.s_addr == m_group.sin_addr.S_un.S_addr )
                {
                    m_response = GroupQuery;
                }
                else if( igmphdr->igmp_group.s_addr == 0 )
                {
                    m_response = GeneralQuery;
                }

                if( !waitForMultipleQuery )
                {
                    break;
                }
            }
        }

        waitInterval = timeoutMs - ( GetTickCount() - timestamp );
        
    }
}


bool MD_Igmp::IsResponseOfType( QueryType type )
{
    return ( m_response == type );
}


void MD_Igmp::SendMembershipReport2( sockaddr_in& group )
{
	IGMPHeader header;

	memset( &header, 0, sizeof( IGMPHeader ) );

	header.igmp_type = IGMP_V2_MEMBERSHIP_REPORT;

	header.igmp_group = group.sin_addr;

	header.igmp_cksum = m_socket.CalculateCheckSum( (short*)&header, sizeof( IGMPHeader ) );

	m_socket.SendTo( (char*)&header, sizeof(IGMPHeader), &group );
}

void MD_Igmp::SendLeaveGroup( sockaddr_in& group )
{
	IGMPHeader header;

	memset( &header, 0, sizeof( IGMPHeader ) );

	header.igmp_type = IGMP_V2_LEAVE_GROUP;

	header.igmp_group = group.sin_addr;

	header.igmp_cksum = m_socket.CalculateCheckSum( (short*)&header, sizeof( IGMPHeader ) );

	sockaddr_in  sendToAddress;
    memset( &sendToAddress, 0, sizeof(sockaddr_in) );

    sendToAddress.sin_addr.s_addr = inet_addr( "224.0.0.2" );
    sendToAddress.sin_family = AF_INET;
	
	m_socket.SendTo( (char*)&header, sizeof(IGMPHeader), &sendToAddress );
}

void MD_Igmp::SendMembershipReport3( sockaddr_in&      group, 
									 sockaddr_in&      source,
									 IGMPV3_RecordType type )
{
	char  buffer[sizeof(IGMPV3MembershipReport) + sizeof(IGMPV3GroupRecord) + sizeof(in_addr)];
	
	memset( buffer, 0, ( sizeof(IGMPV3MembershipReport) + sizeof(IGMPV3GroupRecord) + sizeof(in_addr) ) );
	
	IGMPV3MembershipReport* pHeader = (IGMPV3MembershipReport*)buffer;

	pHeader->igmp_type = IGMP_V3_MEMBERSHIP_REPORT;

	pHeader->igmp_grouprecord_count= htons( 1 );

	IGMPV3GroupRecord* pRecord = (IGMPV3GroupRecord*)(buffer + sizeof(IGMPV3MembershipReport));
	
	pRecord->igmp_record_type = type;

	pRecord->igmp_group = group.sin_addr;

	pRecord->igmp_source_count = htons( 1 );

	memcpy( buffer + sizeof(IGMPV3MembershipReport) + sizeof(IGMPV3GroupRecord),
		    &(source.sin_addr), 
			sizeof(in_addr) );

	pHeader->igmp_cksum = m_socket.CalculateCheckSum( (short*)buffer, 
		                                               sizeof(IGMPV3MembershipReport) +
													   sizeof(IGMPV3GroupRecord) +
													   sizeof(in_addr) );

	sockaddr_in  sendToAddress;
    memset( &sendToAddress, 0, sizeof(sockaddr_in) );

    sendToAddress.sin_addr.s_addr = inet_addr( "224.0.0.22" );
    sendToAddress.sin_family = AF_INET;
	
	m_socket.SendTo( buffer, 
		             sizeof(IGMPV3MembershipReport) + sizeof(IGMPV3GroupRecord) + sizeof(in_addr), 
					 &sendToAddress );
}

void MD_Igmp::JoinGroup( std::string& group )
{
    ip_mreq  mreq;
    mreq.imr_multiaddr.s_addr = inet_addr( group.c_str() );
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);

    m_socket.SetOption( IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*)&mreq, sizeof(ip_mreq) );
}

void MD_Igmp::LeaveGroup( std::string& group )
{
    ip_mreq  mreq;
    mreq.imr_multiaddr.s_addr = inet_addr( group.c_str() );
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);

    m_socket.SetOption( IPPROTO_IP, IP_DROP_MEMBERSHIP, (char*)&mreq, sizeof(ip_mreq) );
}
