/*
Copyright  2005 The Regents of the University of California.
    All Rights Reserved.

    Permission to use, copy, modify, and distribute this software and its
    documentation for educational, research and non-profit purposes,
    without fee, and without a written agreement is hereby granted,
    provided that the above copyright notice, this paragraph and the
    following three paragraphs appear in all copies.

    Permission to incorporate this software into commercial products may
    be obtained by contacting the University of California. For
    information about obtaining such a license contact:

    Chrisanna Waldrop
    Copyright Officer
    805-893-7773
    waldrop@research.ucsb.edu

    This software program and documentation are copyrighted by The Regents
    of the University of California. The software program and
    documentation are supplied "as is", without any accompanying services
    from The Regents. The Regents does not warrant that the operation of
    the program will be uninterrupted or error-free. The end-user
    understands that the program was developed for research purposes and
    is advised not to rely exclusively on the program for any reason.

    IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
    FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
    INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
    ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
    ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. THE UNIVERSITY OF
    CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
    BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE
    MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

#include <Socket.h>

#include <mstcpip.h>
#include <ws2tcpip.h>

MD_Socket::MD_Socket( int type, int protocol, int localAddress, short localport )
{
    m_socket = socket( AF_INET, type, protocol );

    if( m_socket == INVALID_SOCKET )
    {
        throw MD_SocketException( WSAGetLastError(), "socket" );
    }

	sockaddr_in local;

	memset( &local, 0, sizeof(sockaddr_in) );

	local.sin_addr.s_addr = htonl(localAddress);
	local.sin_family = AF_INET;
	local.sin_port = htons(localport);

	Bind( &local );
}


MD_Socket::MD_Socket( int protocol )
{
    m_socket = socket( AF_INET, SOCK_RAW, protocol );

    if( m_socket == INVALID_SOCKET )
    {
        throw MD_RawSocketException( WSAGetLastError() );
    }
}


MD_Socket::~MD_Socket()
{
    if( m_socket != INVALID_SOCKET )
    {
        closesocket( m_socket );
    }
}


int MD_Socket::SendTo( char* buffer, int size, sockaddr_in* toAddr )
{
    int retcode = 0;

    if( m_socket != INVALID_SOCKET )
    {
        retcode = sendto( m_socket, buffer, size, 0, (sockaddr*)toAddr, sizeof(sockaddr_in) );
 
        if( retcode == SOCKET_ERROR )
        {
            throw MD_SocketException( WSAGetLastError(), "sendTo" );
        }
    }

    return retcode;
}

int MD_Socket::ReceiveFrom( char* buffer, int len, sockaddr_in* fromAddr )
{
    int retcode = 0;

    if( m_socket != INVALID_SOCKET )
    {
        int size = sizeof( sockaddr_in );
        retcode = recvfrom( m_socket, buffer, len, 0, (sockaddr*)fromAddr, &size );

        if( retcode == SOCKET_ERROR )
        {
            throw MD_SocketException( WSAGetLastError(), "recvFrom" );
        }
    }

    return retcode;
}


bool MD_Socket::IsDataAvailableForRead(int timeoutMs /* = 0 */ )
{
    bool retval = false;

    if( m_socket != INVALID_SOCKET )
    {
        fd_set fds;
        FD_ZERO( &fds );

        FD_SET( m_socket, &fds );

        div_t divResult = div( timeoutMs, 1000 );
        timeval timeout = { divResult.quot, divResult.rem * 1000 };

        int retCode = select( 0, &fds, NULL, NULL, timeoutMs == INFINITE ? NULL : &timeout );

        if( retCode == SOCKET_ERROR )
        {
            throw MD_SocketException( WSAGetLastError(), "select" );
        }
        else if( retCode != 0 ) //timeout
        {
            retval = true;
        }
    }

    return retval;
}

    
void MD_Socket::SetOption( int level, int optname, char* optionValue, int optionSize )
{
    if( m_socket != INVALID_SOCKET )
    {
        int retcode = setsockopt( m_socket, level, optname, optionValue, optionSize );

        if( retcode == SOCKET_ERROR )
        {
            throw MD_SocketException( WSAGetLastError(), "SetOption" );
        }
    }
}

void MD_Socket::Bind( sockaddr_in* name )
{
    if( m_socket != INVALID_SOCKET )
    {
        int retcode = bind( m_socket, (sockaddr*)name, sizeof(sockaddr_in) );

        if( retcode == SOCKET_ERROR )
        {
            throw MD_SocketException( WSAGetLastError(), "Bind" );
        }
    }
}


void MD_Socket::Bind( int protocol )
{
    if( m_socket != INVALID_SOCKET )
    {
        addrinfo   hints;
		addrinfo*  res = NULL;

		memset(&hints, 0, sizeof(hints));
		hints.ai_flags  = AI_PASSIVE;
		hints.ai_family = AF_INET;
		hints.ai_socktype = SOCK_RAW;
		hints.ai_protocol = protocol;

		int retCode = 0;

		char hostname[1024];
		memset( hostname, 0, 1024 );

		retCode = gethostname( hostname, 1024 );

		if( retCode == SOCKET_ERROR )
		{
			throw MD_SocketException( WSAGetLastError(), "gethostname" );
		}

		retCode = getaddrinfo( hostname, "0", &hints, &res );
           
		if( retCode != 0 )
		{
			throw MD_SocketException( retCode, "getaddrinfo" );
		}

		retCode = bind( m_socket, res->ai_addr, sizeof(sockaddr_in) );

		freeaddrinfo( res );

        if( retCode == SOCKET_ERROR )
        {
            throw MD_SocketException( WSAGetLastError(), "Bind" );
        }
    }
}


short MD_Socket::CalculateCheckSum( short* buffer, int size)
{
    unsigned long cksum = 0;
    
    while (size > 1) {
        cksum += *buffer++;
        size -= sizeof(short);
    }
    if (size) {
        cksum += *(char*)buffer;
    }

    cksum = (cksum >> 16) + (cksum & 0xffff);
    cksum += (cksum >> 16);
    
    return (short)(~cksum);
}


void MD_Socket::SetIoctlForIGMP()
{
	unsigned int optval = 1;
	DWORD        bytes = 0;

    int retCode = WSAIoctl( m_socket, SIO_RCVALL_IGMPMCAST, &optval, sizeof(optval), NULL, 0, &bytes, NULL, NULL);
    
	if (retCode == SOCKET_ERROR)
	{
        throw MD_SocketException( WSAGetLastError(), "WSAIoctl" );
    }
}

void MD_Socket::SetIoctlForAllIP()
{
	unsigned int optval = 1;
	DWORD        bytes = 0;

    int retCode = WSAIoctl( m_socket, SIO_RCVALL, &optval, sizeof(optval), NULL, 0, &bytes, NULL, NULL);
    
	if (retCode == SOCKET_ERROR)
	{
        throw MD_SocketException( WSAGetLastError(), "WSAIoctl" );
    }
}


MD_WinsockInit::MD_WinsockInit()
{
	WSADATA wsaData;

    int err = WSAStartup( MAKEWORD( 2, 2 ), &wsaData );
    
    if ( err != 0 ) 
    {
        throw MD_SocketException( WSAGetLastError(), "WSAStartup" );
    }
}

MD_WinsockInit::~MD_WinsockInit()
{
	int err = WSACleanup();

	if ( err != 0 ) 
    {
        throw MD_SocketException( WSAGetLastError(), "WSACleanup" );
    }
}